from importlib_metadata import version  # type: ignore

__title__ = "snyk"
__description__ = "An API client for the Snyk API."
__url__ = "https://snyk.docs.apiary.io"
__license__ = "MIT"
__version__ = version("pysnyk")
